from typing import Any, MutableMapping

from pybind11.setup_helpers import Pybind11Extension, build_ext

ext_modules = [
    Pybind11Extension(
        name="slumba.cslumba",
        sources=["slumba/cslumba.cc"],
        libraries=["sqlite3"],
    )
]


def build(setup_kwargs: MutableMapping[str, Any]) -> None:
    """This function is mandatory in order to build the extensions."""

    setup_kwargs.update(
        {"ext_modules": ext_modules, "cmdclass": {"build_ext": build_ext}}
    )



from numba import extending, types
from numba.core import cgutils, imputils
from numba.core.typing import ctypes_utils

from slumba.cslumba import SQLITE_NULL
from slumba.sqlite import RESULT_SETTERS, VALUE_EXTRACTORS, sqlite3_value_type


@extending.intrinsic
def unsafe_cast(typingctx, src, dst):
    """Cast a void pointer to a jitclass.

    Parameters
    ----------
    typingctx
    src
        A pointer to cast to type `dst`.
    dst
        A type to cast from `src` to.

    Raises
    ------
    TypeError
        If `src` is not a raw pointer or integer or `dst` is not a jitclass
        type.

    Returns
    -------
    """
    if isinstance(src, (types.RawPointer, types.Integer)) and isinstance(
        dst, types.ClassType
    ):
        inst_typ = dst.instance_type
        sig = inst_typ(types.voidptr, dst)

        def codegen(context, builder, signature, args):
            ptr, _ = args
            alloc_type = context.get_data_type(inst_typ.get_data_type())

            inst_struct = context.make_helper(builder, inst_typ)

            # Set meminfo to correctly typed NULL value
            #
            # If you don't set this attribute to a NULL value, then numba
            # thinks it owns the memory, when in fact SQLite is the owner.
            inst_struct.meminfo = cgutils.get_null_value(inst_struct.meminfo.type)

            # Set data from the given pointer
            inst_struct.data = builder.bitcast(ptr, alloc_type.as_pointer())
            return inst_struct._getvalue()

        return sig, codegen
    else:
        raise TypeError(f"Unable to cast pointer type {src} to class type {dst}")


@extending.intrinsic
def make_arg_tuple(typingctx, func, argv):
    (func_type,), _ = func.get_call_signatures()
    first_arg, *_ = args = func_type.args

    # skip the first argument if `func` is a method call
    first_argument_position = int(isinstance(first_arg, types.ClassInstanceType))
    argtypes = args[first_argument_position:]
    tuple_type = types.Tuple(argtypes)
    sig = tuple_type(func, types.CPointer(types.voidptr))

    def codegen(context, builder, signature, args):
        _, argv = args
        converted_args = []
        for i, argtype in enumerate(argtypes):
            # get a pointer to the ith argument
            element_pointer = cgutils.gep(builder, argv, i, inbounds=True)

            # deref that pointer
            element = builder.load(element_pointer)

            # check for null values #
            # get a pointer to the sqlite3_value_type C function
            sqlite3_value_type_numba = context.get_constant_generic(
                builder,
                ctypes_utils.make_function_type(sqlite3_value_type),
                sqlite3_value_type,
            )
            value_type = builder.call(sqlite3_value_type_numba, [element])

            # make the SQLITE_NULL value type constant available
            sqlite_null = context.get_constant(types.int32, SQLITE_NULL)

            # check whether the value is equal to SQLITE_NULL
            is_null = cgutils.is_true(
                builder, builder.icmp_signed("==", value_type, sqlite_null)
            )

            # setup value extraction #
            # get the appropriate ctypes extraction routine
            ctypes_function = VALUE_EXTRACTORS[argtype]

            # create a numba function type for the converter
            converter = ctypes_utils.make_function_type(ctypes_function)

            # get the function pointer instruction out
            fn = context.get_constant_generic(builder, converter, ctypes_function)

            # if the argument is an optional type then pull out the underlying
            # type and make an optional value with it
            #
            # otherwise the raw value is the argument
            raw = builder.call(fn, [element])
            if isinstance(argtype, types.Optional):
                underlying_type = getattr(argtype, "type", argtype)

                # make an optional none if the value is null, otherwise
                # make an optional value from the raw
                instr = builder.select(
                    is_null,
                    context.make_optional_none(builder, underlying_type),
                    context.make_optional_value(builder, underlying_type, raw),
                )
            else:
                # TODO: should check if a value is null and raise an error if
                # it is
                instr = raw

            # collect the value into an argument list used to build the tuple
            converted_args.append(instr)

        # construct a tuple (fixed length and known types) similar to tuples in
        # statically typed languages
        res = context.make_tuple(builder, tuple_type, converted_args)
        return imputils.impl_ret_borrowed(context, builder, tuple_type, res)

    return sig, codegen


@extending.intrinsic
def get_sqlite3_result_function(typingctx, value_type):
    underlying_type = getattr(value_type, "type", value_type)
    func_type = types.void(types.voidptr, underlying_type)

    external_function_pointer = types.ExternalFunctionPointer(
        func_type, ctypes_utils.get_pointer
    )
    sig = external_function_pointer(underlying_type)

    def codegen(context, builder, signature, args):
        # get the appropriate ctypes extraction routine
        ctypes_function = RESULT_SETTERS[value_type]

        # create a numba function type for the converter
        converter = ctypes_utils.make_function_type(ctypes_function)

        # get the function pointer instruction out
        fn = context.get_constant_generic(builder, converter, ctypes_function)
        return fn

    return sig, codegen


@extending.intrinsic
def sizeof(typingctx, src):
    if isinstance(src, types.ClassType):
        sig = types.int64(src)

        def codegen(context, builder, signature, args):
            data_type = context.get_data_type(src.instance_type)
            size_of_data_type = context.get_abi_sizeof(data_type)
            return context.get_constant(sig.return_type, size_of_data_type)

        return sig, codegen
    else:
        raise TypeError("Cannot get sizeof non jitclass")


def generate_null_checker(func):
    def null_pointer_checker(typingctx, src):
        if isinstance(src, types.ClassInstanceType):
            sig = types.boolean(src)

            def codegen(context, builder, signature, args):
                (instance,) = args

                # TODO: probably a more general way to do this
                second_element = builder.extract_value(instance, [1])
                result = func(builder, second_element)
                return result

            return sig, codegen
        else:
            raise TypeError("Cannot check null pointer status of a non-jitclass type")

    return null_pointer_checker


is_null_pointer = extending.intrinsic(generate_null_checker(cgutils.is_null))
is_not_null_pointer = extending.intrinsic(generate_null_checker(cgutils.is_not_null))



from typing import Callable, Type

from numba import cfunc, void
from numba.core.typing import Signature
from numba.types import CPointer, intc, voidptr

from slumba.numbaext import (
    get_sqlite3_result_function,
    is_not_null_pointer,
    make_arg_tuple,
    sizeof,
    unsafe_cast,
)
from slumba.sqlite import sqlite3_aggregate_context, sqlite3_result_null


def sqlite_udaf(signature: Signature) -> Callable[[Type], Type]:
    """Define a custom aggregate function.

    Parameters
    ----------
    signature
        A numba signature.

    """

    def cls_wrapper(cls: Type) -> Type:
        class_type = cls.class_type
        instance_type = class_type.instance_type

        step_func = class_type.jit_methods["step"]
        step_signature = void(instance_type, *signature.args)
        step_func.compile(step_signature)

        @cfunc(void(voidptr, intc, CPointer(voidptr)))
        def step(ctx, argc, argv):  # pragma: no cover
            raw_pointer = sqlite3_aggregate_context(ctx, sizeof(cls))
            agg_ctx = unsafe_cast(raw_pointer, cls)
            if is_not_null_pointer(agg_ctx):
                args = make_arg_tuple(step_func, argv)
                agg_ctx.step(*args)

        finalize_func = class_type.jit_methods["finalize"]
        finalize_signature: Signature = signature.return_type(instance_type)
        finalize_func.compile(finalize_signature)

        @cfunc(void(voidptr))
        def finalize(ctx):  # pragma: no cover
            raw_pointer = sqlite3_aggregate_context(ctx, sizeof(cls))
            agg_ctx = unsafe_cast(raw_pointer, cls)
            if is_not_null_pointer(agg_ctx):
                result = agg_ctx.finalize()
                if result is None:
                    sqlite3_result_null(ctx)
                else:
                    result_setter = get_sqlite3_result_function(result)
                    result_setter(ctx, result)

        try:
            value_func = class_type.jit_methods["value"]
        except KeyError:
            has_value_func = False
        else:
            has_value_func = True

        try:
            inverse_func = class_type.jit_methods["inverse"]
        except KeyError:
            has_inverse_func = False
        else:
            has_inverse_func = True

        is_window_function = has_value_func and has_inverse_func

        if is_window_function:
            # aggregates can always return a NULL value
            value_signature = signature.return_type(instance_type)
            value_func.compile(value_signature)

            @cfunc(void(voidptr))
            def value(ctx):  # pragma: no cover
                raw_pointer = sqlite3_aggregate_context(ctx, sizeof(cls))
                agg_ctx = unsafe_cast(raw_pointer, cls)
                if is_not_null_pointer(agg_ctx):
                    result = agg_ctx.value()
                    if result is None:
                        sqlite3_result_null(ctx)
                    else:
                        result_setter = get_sqlite3_result_function(result)
                        result_setter(ctx, result)

            inverse_signature: Signature = step_signature
            inverse_func.compile(inverse_signature)

            @cfunc(void(voidptr, intc, CPointer(voidptr)))
            def inverse(ctx, argc, argv):  # pragma: no cover
                raw_pointer = sqlite3_aggregate_context(ctx, sizeof(cls))
                agg_ctx = unsafe_cast(raw_pointer, cls)
                if is_not_null_pointer(agg_ctx):
                    agg_ctx.inverse(*make_arg_tuple(inverse_func, argv))

        cls.step.address = step.address
        cls.finalize.address = finalize.address

        if is_window_function:
            cls.value.address = value.address
            cls.inverse.address = inverse.address
        return cls

    return cls_wrapper



from typing import Any, Callable

from numba import cfunc, njit
from numba.core.ccallback import CFunc
from numba.core.typing import Signature
from numba.types import CPointer, intc, void, voidptr

from slumba.numbaext import get_sqlite3_result_function, make_arg_tuple
from slumba.sqlite import sqlite3_result_null


def sqlite_udf(
    signature: Signature, nogil: bool = True, **njit_kwargs: Any
) -> Callable[[Callable], CFunc]:
    """Define a custom scalar function.

    Parameters
    ----------
    signature
        A numba signature.
    nogil
        Whether to release the GIL.
    nijt_kwargs
        Any additional keyword arguments supported by numba's jit decorator.

    """

    def wrapper(func: Callable) -> CFunc:
        compiled_func = njit(signature, nogil=nogil, **njit_kwargs)(func)

        @cfunc(void(voidptr, intc, CPointer(voidptr)))
        def scalar(ctx, argc, argv):  # pragma: no cover
            args = make_arg_tuple(compiled_func, argv)
            result = compiled_func(*args)
            if result is None:
                sqlite3_result_null(ctx)
            else:
                result_setter = get_sqlite3_result_function(result)
                result_setter(ctx, result)

        return scalar

    return wrapper



from ctypes import (
    CDLL,
    CFUNCTYPE,
    POINTER,
    c_char_p,
    c_double,
    c_int,
    c_int64,
    c_void_p,
)
from ctypes.util import find_library
from typing import Optional

from numba import float64, int32, int64, optional

from .cslumba import SQLITE_DETERMINISTIC, SQLITE_OK, SQLITE_UTF8  # noqa: F401

sqlite3_path: Optional[str] = find_library("sqlite3")
if sqlite3_path is None:  # pragma: no cover
    raise RuntimeError("Unable to find sqlite3 library")

libsqlite3 = CDLL(sqlite3_path)

sqlite3_aggregate_context = libsqlite3.sqlite3_aggregate_context
sqlite3_aggregate_context.argtypes = c_void_p, c_int
sqlite3_aggregate_context.restype = c_void_p

sqlite3_result_double = libsqlite3.sqlite3_result_double
sqlite3_result_int64 = libsqlite3.sqlite3_result_int64
sqlite3_result_int = libsqlite3.sqlite3_result_int
sqlite3_result_null = libsqlite3.sqlite3_result_null

sqlite3_result_double.argtypes = c_void_p, c_double
sqlite3_result_double.restype = None

sqlite3_result_int64.argtypes = c_void_p, c_int64
sqlite3_result_int64.restype = None

sqlite3_result_int.argtypes = c_void_p, c_int
sqlite3_result_int.restype = None

sqlite3_result_null.argtypes = (c_void_p,)
sqlite3_result_null.restype = None

scalarfunc = CFUNCTYPE(None, c_void_p, c_int, POINTER(c_void_p))
stepfunc = CFUNCTYPE(None, c_void_p, c_int, POINTER(c_void_p))
finalizefunc = CFUNCTYPE(None, c_void_p)
valuefunc = CFUNCTYPE(None, c_void_p)
inversefunc = CFUNCTYPE(None, c_void_p, c_int, POINTER(c_void_p))

destroyfunc = CFUNCTYPE(None, c_void_p)

sqlite3_create_function = libsqlite3.sqlite3_create_function
sqlite3_create_function.restype = c_int
sqlite3_create_function.argtypes = (
    c_void_p,
    c_char_p,
    c_int,
    c_int,
    c_void_p,
    scalarfunc,
    stepfunc,
    finalizefunc,
)

try:
    sqlite3_create_window_function = libsqlite3.sqlite3_create_window_function
except AttributeError:  # pragma: no cover
    pass
else:
    sqlite3_create_window_function.restype = c_int
    sqlite3_create_window_function.argtypes = (
        c_void_p,
        c_char_p,
        c_int,
        c_int,
        c_void_p,
        stepfunc,
        finalizefunc,
        valuefunc,
        inversefunc,
        destroyfunc,
    )


RESULT_SETTERS = {
    optional(float64): sqlite3_result_double,
    optional(int64): sqlite3_result_int64,
    optional(int32): sqlite3_result_int,
    float64: sqlite3_result_double,
    int64: sqlite3_result_int64,
    int32: sqlite3_result_int,
}


value_methods = {
    "double": c_double,
    "int": c_int,
    "int64": c_int64,
    "type": c_int,
}


def add_value_method(typename: str, restype):
    method = getattr(libsqlite3, f"sqlite3_value_{typename}")
    method.argtypes = (c_void_p,)
    method.restype = restype
    return method


VALUE_EXTRACTORS = {
    optional(float64): add_value_method("double", c_double),
    optional(int64): add_value_method("int64", c_int64),
    optional(int32): add_value_method("int", c_int),
    float64: add_value_method("double", c_double),
    int64: add_value_method("int64", c_int64),
    int32: add_value_method("int", c_int),
}

sqlite3_value_type = libsqlite3.sqlite3_value_type
sqlite3_value_type.argtypes = (c_void_p,)
sqlite3_restype = c_int

_sqlite3_errmsg = libsqlite3.sqlite3_errmsg
_sqlite3_errmsg.argtypes = (c_void_p,)
_sqlite3_errmsg.restype = c_char_p


def sqlite3_errmsg(db):
    return _sqlite3_errmsg(db).decode("utf8")



import sqlite3
from typing import Callable

from numba.types import ClassType

from .aggregate import sqlite_udaf
from .cslumba import get_sqlite_db
from .scalar import sqlite_udf
from .sqlite import (
    SQLITE_DETERMINISTIC,
    SQLITE_OK,
    SQLITE_UTF8,
    destroyfunc,
    finalizefunc,
    inversefunc,
    scalarfunc,
    sqlite3_create_function,
    sqlite3_create_window_function,
    sqlite3_errmsg,
    stepfunc,
    valuefunc,
)

__all__ = (
    "create_function",
    "create_aggregate",
    "sqlite_udf",
    "sqlite_udaf",
)


def create_function(
    con: sqlite3.Connection,
    name: str,
    num_params: int,
    func: Callable,
    deterministic: bool = False,
) -> None:
    """Register a UDF with name `name` with the SQLite connection `con`.

    Parameters
    ----------
    con : sqlite3.Connection
        A connection to a SQLite database
    name : str
        The name of this function in the database, given as a UTF-8 encoded
        string
    num_params : int
        The number of arguments this function takes
    func : cfunc
        The sqlite_udf-decorated function to register
    deterministic : bool
        True if this function returns the same output given the same input.
        Most functions are deterministic.

    """
    sqlite_db = get_sqlite_db(con)
    if (
        sqlite3_create_function(
            sqlite_db,
            name.encode("utf8"),
            num_params,
            SQLITE_UTF8 | (SQLITE_DETERMINISTIC if deterministic else 0),
            None,
            scalarfunc(getattr(func, "address")),
            stepfunc(0),
            finalizefunc(0),
        )
        != SQLITE_OK
    ):
        raise sqlite3.OperationalError(sqlite3_errmsg(sqlite_db))


def create_aggregate(
    con: sqlite3.Connection,
    name: str,
    num_params: int,
    aggregate_class: ClassType,
    deterministic: bool = False,
) -> None:
    """Register an aggregate named `name` with the SQLite connection `con`.

    Parameters
    ----------
    con : sqlite3.Connection
        A connection to a SQLite database
    name : str
        The name of this function in the database, given as a UTF-8 encoded
        string
    num_params : int
        The number of arguments this function takes
    aggregate_class : JitClass
       This class must be decorated with @sqlite_udaf for this function to
       work. If this class has `value` and `inverse` attributes, it will be
       registered as a window function. Window functions can also be used as
       standard aggregate functions.
    deterministic : bool
        True if this function returns the same output given the same input.
        Most functions are deterministic.

    """
    namebytes = name.encode("utf8")
    sqlite_db = get_sqlite_db(con)
    if hasattr(aggregate_class, "value") and hasattr(aggregate_class, "inverse"):
        rc = sqlite3_create_window_function(
            sqlite_db,
            namebytes,
            num_params,
            SQLITE_UTF8 | (SQLITE_DETERMINISTIC if deterministic else 0),
            None,
            stepfunc(aggregate_class.step.address),
            finalizefunc(aggregate_class.finalize.address),
            valuefunc(aggregate_class.value.address),
            inversefunc(aggregate_class.inverse.address),
            destroyfunc(0),
        )
    else:
        rc = sqlite3_create_function(
            sqlite_db,
            namebytes,
            num_params,
            SQLITE_UTF8 | (SQLITE_DETERMINISTIC if deterministic else 0),
            None,
            scalarfunc(0),
            stepfunc(aggregate_class.step.address),
            finalizefunc(aggregate_class.finalize.address),
        )
    if rc != SQLITE_OK:
        raise sqlite3.OperationalError(sqlite3_errmsg(sqlite_db))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['slumba']

package_data = \
{'': ['*']}

install_requires = \
['numba==0.53.1']

setup_kwargs = {
    'name': 'slumba',
    'version': '1.0.6',
    'description': 'JITted SQLite user-defined functions and aggregates',
    'long_description': '# Put some Numba in your SQLite\n\n## Fair Warning\n\nThis library does unsafe things like pass around function pointer addresses\nas integers.  **Use at your own risk**.\n\nIf you\'re unfamiliar with why passing function pointers\' addresses around as\nintegers might be unsafe, then you shouldn\'t use this library.\n\n## Requirements\n\n* Python >=3.7\n* `numba`\n\nUse `nix-shell` from the repository to avoid dependency hell.\n\n## Installation\n* `poetry install`\n\n## Examples\n\n### Scalar Functions\n\nThese are almost the same as decorating a Python function with\n`numba.jit`. In the case of `sqlite_udf` a signature is required.\n\n```python\nfrom slumba import sqlite_udf\nfrom numba import int64\n\n\n@sqlite_udf(optional(int64)(optional(int64)))\ndef add_one(x):\n    """Add one to `x` if `x` is not NULL."""\n\n    if x is not None:\n        return x + 1\n    return None\n```\n\n\n### Aggregate Functions\n\nThese follow the API of the Python standard library\'s\n`sqlite3.Connection.create_aggregate` method. The difference with slumba\naggregates is that they require two decorators: `numba.experimental.jit_class` and\n`slumba.sqlite_udaf`. Let\'s define the `avg` (arithmetic mean) function for\n64-bit floating point numbers.\n\n```python\nfrom numba import int64, float64\nfrom numba.experimental import jit_class\nfrom slumba import sqlite_udaf\n\n\n@sqlite_udaf(optional(float64)(optional(float64)))\n@jit_class(dict(total=float64, count=int64))\nclass Avg:\n    def __init__(self):\n        self.total = 0.0\n        self.count = 0\n\n    def step(self, value):\n        if value is not None:\n            self.total += value\n            self.count += 1\n\n    def finalize(self):\n        if not self.count:\n            return None\n        return self.total / self.count\n```\n\n### Window Functions\n\nYou can also define window functions for use with SQLite\'s `OVER` construct:\n\n```python\n@sqlite_udaf(optional(float64)(optional(float64)))\n@jitclass(dict(total=float64, count=int64))\nclass WinAvg:  # pragma: no cover\n    def __init__(self):\n        self.total = 0.0\n        self.count = 0\n\n    def step(self, value):\n        if value is not None:\n            self.total += value\n            self.count += 1\n\n    def finalize(self):\n        count = self.count\n        if count:\n            return self.total / count\n        return None\n\n    def value(self):\n        return self.finalize()\n\n    def inverse(self, value):\n        if value is not None:\n            self.total -= value\n            self.count -= 1\n```\n',
    'author': 'Phillip Cloud',
    'author_email': '417981+cpcloud@users.noreply.github.com',
    'maintainer': 'Phillip Cloud',
    'maintainer_email': '417981+cpcloud@users.noreply.github.com',
    'url': 'https://github.com/cpcloud/slumba',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.8,<3.10',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)
